﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using RemoteDatabase;

namespace RemoteProxy {

  // class RProxy impersonates the remote database and talks to the remote bank account:
  // (i) It holds the networking code that does remote procedure calls (RPCs) 
  //   to the remote database.
  // (ii) RProxy also holds code for doing account-balance lookups and account withdrawals.
  //   This will execute slowly.

  public class RProxy : IDB, IAcct { // This one class handles logins and account lookups.
                                     //  (We will fix this so that account lookups are 
                                     // handled by a separate remote proxy for the account.)

    private AccountDB remoteDatabase;  // handle to the bank's database that lives far away
    private Account openedAccount;     // handle to the opened bank account
    private string accountid;          // the id of the opened bank account

    public RProxy(AccountDB remote) { 
        remoteDatabase = remote;
        openedAccount = null;
        accountid = null;
    }


    /// This part of RProxy impersonates the remote database:

    // login to bank account  #id:
    // Param: id - the name of the account to login to
    // Returns: a handle that is used by the ATM client to do account lookups and withdrawals
    public IAcct login(string id) {
      // call the remote database via RPC to login and receive in return 
      //   a handle to the remote bank acct:
      openedAccount = (Account)(rpcToDatabase(remoteDatabase.login, id));   //  see method  rpcToDatabase below
      if (openedAccount != null)  {  // login succeeded
          accountid = id; 
          return this;   // see methods  getBalance and withdraw,  coded below....
      }
      else { return null; }  // login failed
    }

    // logout from account #id:
    public object logout(string id) {
      if (accountid != null && accountid == id) {
          rpcToDatabase(remoteDatabase.logout, accountid);
          openedAccount = null; 
      }
      return null;
    }

    /// This part of RProxy does the lookups to the remote bank account --- 
    /// It does RPCs to the account, which is located far away:

    // lookup balance of remote account:
    public int getBalance() {
      int answer = -1;
      if (openedAccount != null) {
          answer = rpcToAccount(openedAccount.getBalance); 
      }
      return answer;
    }

    // do withdrawal of  amt  from remote account and return whether successful:
    public bool withdraw(int amt) {
      bool answer = false;
      if (openedAccount != null) {
        answer = rpcToAccount(openedAccount.withdraw, amt);
      }
      return answer;
    }


    //////////////////////////////////////////////////////////////////////////////////////
    // The following methods are my "simulations" for the networking code needed for RPCs:
    //////////////////////////////////////////////////////////////////////////////////////  

   // code to simulate RPCs to the remote database:

   // The delegates define the forms of method that can be called remotely:
   private delegate object DBmethod1(string arg);
   private delegate void DBmethod2(string arg, int amt);

   // does an RPC for method  m  with argument  arg.  Returns an *object* as the answer.
   // The object can be cast to an int or a string or an Account, or whatever it is!
   private object rpcToDatabase(DBmethod1 m, string arg) {
       wait(10);
       return m(arg);
   }
   // does an RPC for method  m  with arguments  arg,amt.  No answer is returned.
   private void rpcToDatabase(DBmethod2 m, string arg, int amt) {
       wait(10);
       m(arg, amt);
   }

   // code to simulate RPCs to a remote bank account; similar to the code just above.
   private delegate bool Acctmethod1(int arg);
   private bool rpcToAccount(Acctmethod1 m, int arg) {
       wait(10);
       return m(arg);
   }
   delegate int Acctmethod2();
   private int rpcToAccount(Acctmethod2 m) {
       wait(10);
       return m();
   }

   private InfoForm f;      // passive form used to show you when an RPC occurs

   // stand-in method for the networking code that issues a remote proc call (RPC)
   private void wait(int howlong)
   {
       if (f == null) { f = new InfoForm(); }
       f.Show();
       for (int i = 0; i != howlong; i++)
       {
           f.WriteLine(i.ToString());
           Thread.Sleep(600);
       }
       f.Hide();
   }
  
  }
}
